{
  KOLOdbc.pas (c) by Vladimir Kladov, 2005, 2013
  11.03.2005, 27.12.2005, 25.10.2013

  :
    EODBC         -     ODBC
    TODBCDatabase -        ODBC
    TODBCQuery    -         ODBC
}

unit KOLOdbc;

interface

//{$DEFINE KOL_EXCEPTIONS} //uncomment or add to project options
                           // to get exceptions

uses Windows, Classes, {$IFDEF KOL}KOL, KOLmath, err
                       {$ELSE} SysUtils, Forms, Db, math{$ENDIF};


type
  {$IFDEF KOL_EXCEPTIONS}
  EODBC = class( Exception )
  protected
    constructor CreateError( HandleType: SmallInt; fHandle: THandle; ErrCode: SmallInt; const Msg: String );
  public
    constructor CreateEnvError( fHandle: THandle; ErrCode: SmallInt; const Msg: String );
    constructor CreateConnError( fHandle: THandle; ErrCode: SmallInt; const Msg: String );
    constructor CreateStmtError( fHandle: THandle; ErrCode: SmallInt; const Msg: String );
  end;
  {$ENDIF KOL_EXCEPTIONS}

  {$IFDEF KOL}
  PODBCDatabase = ^TODBCDatabase;
  {$ENDIF}
  TODBCDatabase = {$IFDEF KOL} object(TObj)
                  {$ELSE} class( TObject )
                  {$ENDIF}
  protected
    fConnected: Boolean;
    fEnvHandle: THandle;
    fConnHandle: THandle;
    fQueryList: {$IFDEF KOL} PList {$ELSE} TList {$ENDIF};
    FAsync: Boolean;
    function GetAsync: Boolean;
    procedure SetAsync(const Value: Boolean);
    {$IFDEF KOL}
    procedure Init; virtual;
    {$ENDIF}
  public
    ConnectionString: String;
    Error: String;
    property EnvironmentHandle: THandle read fEnvHandle;
    property Connection: THandle read fConnHandle;
    property Connected: Boolean read fConnected;
    function Connect: Boolean;
    function DSNConnect( const DSN, User, Pwd: String ): Boolean;
    procedure Disconnect;
    destructor Destroy; {$IFDEF KOL} virtual {$ELSE} override {$ENDIF};
    {$IFNDEF KOL}
    constructor Create;
    {$ENDIF}
    procedure BeginTrans;
    procedure RollbackTrans;
    procedure CommitTrans;
    property Async: Boolean read GetAsync write SetAsync;
  end;

var DefaultODBCDataSource: {$IFDEF KOL} PODBCDatabase
                           {$ELSE}      TODBCDatabase
                           {$ENDIF};

{$IFDEF KOL}
function NewODBCDatabase: PODBCDatabase;
{$ENDIF}

type
  {$IFDEF KOL}
  PODBCQuery = ^TODBCQuery;
  PSQLStrList = ^TSQLStrList;
  TSQLStrList = object( TStrList )
    FOwnerODBCQuery: PODBCQuery;
  end;
  {$ELSE}
  TODBCQuery = class;

  TSQLStringList = class( TStringList )
  protected
    FOwnerODBCQuery: TODBCQuery;
    procedure Changed; override;
  end;
  {$ENDIF}

  {$IFDEF KOL}
  TFieldType = (ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
    ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
    ftBytes, ftVarBytes, ftAutoInc, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
    ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftFixedChar, ftWideString,
    ftLargeint, ftADT, ftArray, ftReference, ftDataSet, ftOraBlob, ftOraClob,
    ftVariant, ftInterface, ftIDispatch, ftGuid);
  {$ENDIF}

  TFieldDescription = packed record
    DataType           : SmallInt;
    ColumnSize         : DWORD;
    DecimalDigits      : SmallInt;
    Nullable           : SmallInt;
    BindPtr  : Pointer;
    StrLenInd: PInteger;
  end;
  PFieldDescription = ^TFieldDescription;

  TODBCQuery = {$IFDEF KOL} object(TObj)
               {$ELSE}      class( TComponent )
               {$ENDIF}
  protected
    function GetColCount: Integer;
    function GetFieldNames(Idx: Integer): String;
    function GetSField(const S: String): String;
    function GetSFieldN(N: Integer): String;
    function GetIField(const S: String): Integer;
    function GetIFieldN(N: Integer): Integer;
    function GetFField(const S: String): Extended;
    function GetFFieldN(N: Integer): Extended;
    function GetDField(const S: String): TDateTime;
    function GetDFieldN(N: Integer): TDateTime;
    function GetNullField(const S: String): Boolean;
    function GetNullFieldN(N: Integer): Boolean;
    function GetFieldTypes(Idx: Integer): TFieldType;
    function GetFieldPrecision(Idx: Integer): Integer;
  protected
    fHandle: Integer;
    fColCount: SmallInt;
    fColumns: {$IFDEF KOL} PStrListEx {$ELSE} TStringList {$ENDIF};
    fBindBuffer: Pointer;
    fEof, fBof: Boolean;
    procedure ClearColumns;
    function ColumnByName( const S: String ): Integer;
    function ColumnByNameMustBe( const S: String ): Integer;
    procedure InternalOpenResultset;
    {$IFDEF KOL}
    procedure Init; virtual;
    {$ENDIF}
  public
    Error: String;
    ODBC: {$IFDEF KOL} PODBCDatabase {$ELSE} TODBCDatabase {$ENDIF};
    SQL: {$IFDEF KOL} PSQLStrList {$ELSE} TSQLStringList {$ENDIF};
    {$IFNDEF KOL}
    constructor Create( AOwner: TComponent ); override;
    {$ENDIF}
    destructor Destroy; {$IFDEF KOL} virtual {$ELSE} override {$ENDIF};
    function ExecSQL: Boolean;
    function Open: Boolean;
    procedure Close;
    property FieldCount: Integer read GetColCount;
    property FieldNames[ Idx: Integer ]: String read GetFieldNames;
    property FieldTypes[ Idx: Integer ]: TFieldType read GetFieldTypes;
    property FieldPRecision[ Idx: Integer ]: Integer read GetFieldPrecision;
    function FindField( const S: String ): Integer;
    property Eof: Boolean read fEof;
    property Bof: Boolean read fBof;
    procedure First;
    procedure Next;
    procedure Prev;
    procedure Last;
    function More: Boolean;
    property SFieldN[ N: Integer ]: String read GetSFieldN;
    property SField[ const S: String ]: String read GetSField;
    property IFieldN[ N: Integer ]: Integer read GetIFieldN;
    property IField[ const S: String ]: Integer read GetIField;
    property FFieldN[ N: Integer ]: Extended read GetFFieldN;
    property FField[ const S: String ]: Extended read GetFField;
    property DFieldN[ N: Integer ]: TDateTime read GetDFieldN;
    property DField[ const S: String ]: TDateTime read GetDField;
    property NullFieldN[ N: Integer ]: Boolean read GetNullFieldN;
    property NullField[ const S: String ]: Boolean read GetNullField;
  end;

{$IFDEF KOL}
function NewODBCQuery: PODBCQuery;
{$ENDIF}

implementation

//-------------------------   API ODBC --------------------------
const ODBC32 = 'ODBC32.dll';
      SQL_OV_ODBC3 = 3;

const
  SQL_CP_ONE_PER_DRIVER = 1;
  SQL_CP_ONE_PER_HENV   = 2;
  SQL_ATTR_ODBC_VERSION	= 200;
  SQL_IS_INTEGER	= -6;
  SQL_ATTR_ASYNC_ENABLE = 4;
  SQL_ASYNC_ENABLE_OFF = 0;
  SQL_NTS = -3;

type SQLResult = Word;
     TSqlDateTimeStamp = packed record
     Year, month, day, hour, minute, second: Word;
     fraction: DWORD; // 1/1000 000 000 sec
     end;
     PSqlDateTimeStamp = ^TSqlDateTimeStamp;

function SqlDateTimeStampToDateTime( TS: PSqlDateTimeStamp ): TDateTime;
var ST: TSystemTime;
begin
  ST.wYear := TS.Year;
  ST.wMonth := TS.month;
  ST.wDay := TS.day;
  ST.wHour := TS.hour;
  ST.wMinute := TS.minute;
  ST.wSecond := TS.second;
  ST.wMilliseconds := TS.fraction div 1000000;
  {$IFDEF KOL}
  SystemTime2DateTime( ST, Result );
  {$ELSE}
  Result := SystemTimeToDateTime( ST );
  {$ENDIF}
end;

function StrToFloat( const S: String ): Extended;
begin
  if Trim( S ) = '' then
    Result := 0
  else
  TRY
    Result :=
      {$IFDEF KOL}
      Str2Double( S );
      {$ELSE}
      SysUtils.StrToFloat(
      StringReplace(
      StringReplace( S, '.', DecimalSeparator, [] ),
                        ',', DecimalSeparator, [] ) );
      {$ENDIF}
  EXCEPT
    Result := 0;
  END;
end;

{$IFDEF KOL}
function IntToStr( i: Integer ): String;
begin
  Result := Int2Str( i );
end;

function StrToInt( const s: String ): Integer;
begin
  Result := Str2Int( s );
end;
{$ENDIF}

function SQLSetEnvAttr(
  EnvironmentHandle    : THandle;
  Attribute            : Integer;
  ValuePtr             : Pointer;
  StringLength         : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLSetEnvAttr';

const
  SQL_HANDLE_ENV  = 1;
  SQL_HANDLE_DBC  = 2;
  SQL_HANDLE_STMT = 3;
  SQL_HANDLE_DESC = 4;

function SQLAllocHandle(
  HandleType            : SmallInt;
  InputHandle           : THandle;
  OutputHandlePtr       : PHandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLAllocHandle';

function SQLFreeHandle(
  HandleType           : SmallInt;
  Handle               : THandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLFreeHandle';

function SQLDriverConnect(
  ConnectionHandle        : THandle;
  WindowHandle            : HWnd;
  InConnectionString      : PChar;
  StringLength1           : SmallInt;
  OutConnectionString     : PChar;
  BufferLength            : SmallInt;
  StringLength2Ptr        : PSmallInt;
  DriverCompletion        : Word
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLDriverConnect';

function SQLConnect(
  ConnectionHandle  : THandle;
  ServerName        : PChar;
  NameLength1       : SmallInt;
  UserName          : PChar;
  NameLength2       : SmallInt;
  Authentication    : PChar;
  NameLength3       : SmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLConnect';

function SQLDisconnect(
  ConnectionHandle     : THandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLDisconnect';

const
  SQL_COMMIT   = 0;
  SQL_ROLLBACK = 1;

function SQLSetConnectAttr(
  ConnectionHandle         : THandle;
  Attribute                : Integer;
  ValuePtr                 : Pointer;
  StringLength             : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLSetConnectAttr';

function SQLGetConnectAttr(
     ConnectionHandle      : THandle;
     Attribute             : Integer;
     ValuePtr              : Pointer;
     BufferLength          : Integer;
     StringLengthPtr       : PInteger
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLGetConnectAttr';

function SQLEndTran(
  HandleType        : SmallInt;
  Handle            : THandle;
  CompletionType    : SmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLEndTran';

function SQLGetDiagRec(
  HandleType           : SmallInt;
  Handle               : THandle;
  RecNumber            : Integer;
  Sqlstate             : PChar;
  NativeErrorPtr       : PInteger;
  MessageText          : PChar;
  BufferLength         : SmallInt;
  TextLengthPtr        : PSmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLGetDiagRec';

function SQLExecDirect(
  StatementHandle      : THandle;
  StatementText        : PChar;
  TextLength           : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLExecDirect';

function SQLExecute(
  StatementHandle      : THandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLExecute';

function SQLPrepare(
  StatementHandle      : THandle;
  StatementText        : PChar;
  TextLength           : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLPrepare';

function SQLMoreResults(
  StatementHandle      : THandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLMoreResults';

const
  SQL_ATTR_AUTOCOMMIT = 102;
  SQL_AUTOCOMMIT_OFF  = 0;
  SQL_AUTOCOMMIT_ON   = 1;
  SQL_CLOSE           = 0;

function SQLFreeStmt(
  StatementHandle    : THandle;
  Option             : SmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLFreeStmt';

function SQLNumResultCols(
  StatementHandle         : THandle;
  ColumnCountPtr          : PSmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLNumResultCols';

function SQLDescribeCol(
  StatementHandle       : THandle;
  ColumnNumber          : SmallInt;
  ColumnName            : PChar;
  BufferLength          : SmallInt;
  NameLengthPtr         : PSmallInt;
  DataTypePtr           : PSmallInt;
  ColumnSizePtr         : PDWORD;
  DecimalDigitsPtr      : PSmallInt;
  NullablePtr           : PSmallInt
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLDescribeCol';

const
  SQL_UNKNOWN_TYPE = 0;
  SQL_CHAR         = 1;
  SQL_NUMERIC      = 2;
  SQL_DECIMAL      = 3;
  SQL_INTEGER      = 4;
  SQL_SMALLINT     = 5;
  SQL_FLOAT        = 6;
  SQL_REAL         = 7;
  SQL_DOUBLE       = 8;
  SQL_DATE         = 9;
  SQL_TIME         = 10;
  SQL_TYPE_TIMESTAMP = 93;
  SQL_VARCHAR      = 12;
  SQL_WVARCHAR     = -96;

function SQLBindCol(
  StatementHandle   : THandle;
  ColumnNumber      : Word;
  TargetType        : SmallInt;
  TargetValuePtr    : Pointer;
  BufferLength      : Integer;
  StrLen_or_Ind     : PInteger
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLBindCol';

const
  SQL_ATTR_ROW_BIND_TYPE  = 5;
  SQL_ATTR_ROW_ARRAY_SIZE = 27;

function SQLSetStmtAttr(
  StatementHandle       : THandle;
  Attribute             : Integer;
  ValuePtr              : Pointer;
  StringLength          : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLSetStmtAttr';

const
  SQL_FETCH_NEXT      = 1;
  SQL_FETCH_FIRST     = 2;
  SQL_FETCH_LAST      = 3;
  SQL_FETCH_PRIOR     = 4;
  SQL_FETCH_ABSOLUTE  = 5;
  SQL_FETCH_RELATIVE  = 6;

  SQL_NULL_DATA       = -1;

function SQLFetch(
  StatementHandle : THandle
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLFetch';

function SQLFetchScroll(
  StatementHandle       : THandle;
  FetchOrientation      : SmallInt;
  FetchOffset           : Integer
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLFetchScroll';

function SQLGetData(
  StatementHandle   : THandle;
  ColumnNumber      : SmallInt;
  TargetType        : SmallInt;
  TargetValuePtr    : Pointer;
  BufferLength      : Integer;
  StrLen_or_IndPtr  : PInteger
  ): SQLResult; stdcall;
  external ODBC32 name 'SQLGetData';

{$IFDEF KOL}
function NewODBCDatabase: PODBCDatabase;
begin
  new( Result, Create );
end;
{$ENDIF}

{ TODBCDatabase }

procedure TODBCDatabase.BeginTrans;
begin
  SQLSetConnectAttr( fConnHandle, SQL_ATTR_AUTOCOMMIT,
    Pointer( SQL_AUTOCOMMIT_OFF ), SQL_IS_INTEGER );
end;

procedure TODBCDatabase.CommitTrans;
begin
  SQLEndTran( SQL_HANDLE_DBC, fConnHandle, SQL_COMMIT );
  SQLSetConnectAttr( fConnHandle, SQL_ATTR_AUTOCOMMIT,
    Pointer( SQL_AUTOCOMMIT_ON ), SQL_IS_INTEGER );
end;

function TODBCDatabase.Connect: Boolean;
var OutConnString: array[ 0..1023 ] of Char;
    OutLen: SmallInt;
    R: Word;
begin
  Result := TRUE;
  Error := '';
  if Connected then Exit;
  R := SQLAllocHandle( SQL_HANDLE_DBC, fEnvHandle, @ fConnHandle );
  if R > 1 then
  begin
      Error := '   ';
      {$IFDEF KOL_EXCEPTIONS}
      raise EODBC.CreateEnvError( fEnvHandle, R, Error );
      {$ELSE}
      Result := FALSE;
      Exit;
      {$ENDIF KOL_EXCEPTIONS}
  end;
  R := SQLDriverConnect( fConnHandle,
    {$IFDEF KOL} Applet.Handle {$ELSE} Application.Handle {$ENDIF},
    PChar( ConnectionString ),
    Length( ConnectionString ), @ OutConnString[ 0 ], Sizeof( OutConnString ),
    @ OutLen, 0 );
  if R <= 1 then
  begin
    fConnected := TRUE;
    Async := FAsync;
  end
  else
  begin
    TRY
      Error := '   ';
      {$IFDEF KOL_EXCEPTIONS}
      raise EODBC.CreateConnError( fConnHandle, R, Error );
      {$ELSE}
      Result := FALSE;
      Exit;
      {$ENDIF KOL_EXCEPTIONS}
    FINALLY
      SQLFreeHandle( SQL_HANDLE_DBC, fConnHandle );
      fConnHandle := 0;
    END;
  end;
  Result := fConnected;
end;

{$IFDEF KOL}
procedure TODBCDatabase.Init;
var R: WORD;
begin
  inherited;
  fQueryList := NewList;
  R := SQLAllocHandle( SQL_HANDLE_ENV, 0, @ fEnvHandle );
  if R > 1 then
  begin
    Error := '    ODBC';
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.CreateError( 0, 0, R, Error );
    {$ELSE}
    Exit;
    {$ENDIF KOL_EXCEPTIONS}
  end;
  if DefaultODBCDataSource = nil then
    DefaultODBCDataSource := @Self;
  R := SQLSetEnvAttr( fEnvHandle, SQL_ATTR_ODBC_VERSION,
    Pointer( SQL_OV_ODBC3 ), SQL_IS_INTEGER );
  if R > 1 then
  begin
    Error := '    ODBC 3.0';
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.CreateEnvError( fEnvHandle, R, Error );
    {$ELSE}
    Exit;
    {$ENDIF KOL_EXCEPTIONS}
  end;
end;
{$ELSE}
constructor TODBCDatabase.Create;
var R: WORD;
begin
  inherited;
  fQueryList := TList.Create;
  R := SQLAllocHandle( SQL_HANDLE_ENV, 0, @ fEnvHandle );
  if R > 1 then
    raise EODBC.CreateError( 0, 0, R, '    ODBC' );
  if DefaultODBCDataSource = nil then
    DefaultODBCDataSource := Self;
  R := SQLSetEnvAttr( fEnvHandle, SQL_ATTR_ODBC_VERSION,
    Pointer( SQL_OV_ODBC3 ), SQL_IS_INTEGER );
  if R > 1 then
    raise EODBC.CreateEnvError( fEnvHandle, R, '    ODBC 3.0' );
end;
{$ENDIF}

destructor TODBCDatabase.Destroy;
begin
  if DefaultODBCDataSource = {$IFDEF KOL} @ {$ENDIF} Self then
    DefaultODBCDataSource := nil;
  if Connected then
    Disconnect;
  SQLFreeHandle( SQL_HANDLE_ENV, fEnvHandle );
  fQueryList.Free;
  inherited;
end;

procedure TODBCDatabase.Disconnect;
var i: Integer;
    Q: {$IFDEF KOL} PODBCQuery {$ELSE} TODBCQuery {$ENDIF};
begin
  if not Connected then Exit;
  for i := fQueryList.Count-1 downto 0 do
  begin
    Q := fQueryList.Items[ i ];
    Q.Close;
  end;
  SQLDisconnect( fConnHandle );
  SQLFreeHandle( SQL_HANDLE_DBC, fConnHandle );
  fConnHandle := 0;
  fConnected := FALSE;
end;

function TODBCDatabase.DSNConnect(const DSN, User, Pwd: String): Boolean;
var R: Word;
begin
  Result := TRUE;
  if Connected then Exit;
  R := SQLAllocHandle( SQL_HANDLE_DBC, fEnvHandle, @ fConnHandle );
  if R > 1 then
  begin
    Error := '   ';
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.CreateEnvError( fEnvHandle, R, Error );
    {$ELSE}
    Result := FALSE;
    Exit;
    {$ENDIF KOL_EXCEPTIONS}
  end;
  R := SQLConnect( fConnHandle, PChar( DSN ), Length( DSN ),
    PChar( User ), Length( User ), PChar( Pwd ), Length( Pwd ) );
  if R <= 1 then
  begin
    fConnected := TRUE;
    Async := FAsync;
  end
  else
  begin
    TRY
      Error := '   ';
      {$IFDEF KOL_EXCEPTIONS}
      raise EODBC.CreateConnError( fConnHandle, R, Error );
      {$ELSE}
      Result := FALSE;
      Exit;
      {$ENDIF KOL_EXCEPTIONS}
    FINALLY
      SQLFreeHandle( SQL_HANDLE_DBC, fConnHandle );
      fConnHandle := 0;
    END;
  end;
  Result := fConnected;
end;

function TODBCDatabase.GetAsync: Boolean;
var Value: Integer;
    R: Word;
begin
  Result := FAsync;
  if not Connected then Exit;
  R := SQLGetConnectAttr( fConnHandle, SQL_ATTR_ASYNC_ENABLE, @ Value,
    0, nil );
  if R > 1 then Exit;
  FAsync := Value <> 0;
end;

procedure TODBCDatabase.RollbackTrans;
begin
  SQLEndTran( SQL_HANDLE_DBC, fConnHandle, SQL_ROLLBACK );
  SQLSetConnectAttr( fConnHandle, SQL_ATTR_AUTOCOMMIT,
    Pointer( SQL_AUTOCOMMIT_ON ), SQL_IS_INTEGER );
end;

procedure TODBCDatabase.SetAsync(const Value: Boolean);
var R: Word;
begin
  FAsync := Value;
  if not Connected then Exit;
  Error := '';
  R := SQLSetConnectAttr(fConnHandle, SQL_ATTR_ASYNC_ENABLE,
                  Pointer( FAsync ), SQL_NTS);
  if R > 1 then
  begin
    Error := '     = ' +
      IntToStr( Integer( Value ) );
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.CreateError( SQL_HANDLE_DBC, fConnHandle, R, Error );
    {$ELSE}
    {$ENDIF KOL_EXCEPTIONS}
  end;
end;

{ EODBC }

{$IFDEF KOL_EXCEPTIONS}
constructor EODBC.CreateConnError(fHandle: THandle; ErrCode: SmallInt;
  const Msg: String);
begin
  CreateError( SQL_HANDLE_DBC, fHandle, ErrCode, Msg );
end;

constructor EODBC.CreateEnvError( fHandle: THandle; ErrCode: SmallInt; const Msg: String);
begin
  CreateError( SQL_HANDLE_ENV, fHandle, ErrCode, Msg );
end;

constructor EODBC.CreateError(HandleType: SmallInt; fHandle: THandle;
  ErrCode: SmallInt; const Msg: String);
var ErrStringBuffer: array[ 0..4095 ] of Char;
    ErrStrLen: SmallInt;
    SQLState: array[ 0..5 ] of Char;
    NativeError: Integer;
begin
  ErrStringBuffer[ 0 ] := #0;
  ErrStrLen := 0;
  SQLGetDiagRec( HandleType, fHandle, 1, SQLState, @ NativeError,
    ErrStringBuffer, Sizeof( ErrStringBuffer ), @ ErrStrLen );
  ErrStringBuffer[ ErrStrLen ] := #0;
  SQLState[ 5 ] := #0;
  inherited Create( {$IFDEF KOL}e_custom,{$ENDIF}
            Msg + ',  : ' + IntToStr( ErrCode ) +
            #13#10 + SQLState + ' ' + ErrStringBuffer );
end;

constructor EODBC.CreateStmtError(fHandle: THandle; ErrCode: SmallInt;
  const Msg: String);
begin
  CreateError( SQL_HANDLE_STMT, fHandle, ErrCode, Msg );
end;
{$ENDIF KOL_EXCEPTIONS}

{$IFDEF KOL}
function NewODBCQuery: PODBCQuery;
begin
  new( Result, Create );
end;
{$ENDIF}

{ TODBCQuery }

procedure TODBCQuery.ClearColumns;
var i: Integer;
    FD: PFieldDescription;
begin
  for i := fColumns.Count-1 downto 0 do
  begin
    FD := Pointer( fColumns.Objects[ i ] );
    FreeMem( FD );
  end;
  fColumns.Clear;
  fColCount := 0;
  if fBindBuffer <> nil then
  begin
    FreeMem( fBindBuffer );
    fBindBuffer := nil;
  end;
end;

procedure TODBCQuery.Close;
begin
  if fHandle = 0 then Exit;
  //SQLFreeStmt( fHandle, SQL_CLOSE );
  SQLFreeHandle( SQL_HANDLE_STMT, fHandle );
  fHandle := 0;
  ClearColumns;
end;

function TODBCQuery.ColumnByName(const S: String): Integer;
var i: Integer;
begin
  for i := 0 to fColumns.Count-1 do
  begin
    if AnsiCompareText( S, fColumns{$IFDEF KOL}.Items{$ENDIF}[ i ] ) = 0 then
    begin
      Result := i;
      Exit;
    end;
  end;
  Result := -1;
end;

function TODBCQuery.ColumnByNameMustBe(const S: String): Integer;
begin
  Result := ColumnByName( S );
  if Result < 0 then
  begin
    Error := ' ' + S + '     ';
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.Create( {$IFDEF KOL}e_custom,{$ENDIF} Error );
    {$ELSE}
    Result := -1;
    {$ENDIF}
  end;
end;

{$IFDEF KOL}
procedure TODBCQuery.Init;
begin
  inherited;
  new( SQL, Create );
  SQL.FOwnerODBCQuery := @ Self;
  ODBC := DefaultODBCDataSource;
  if ODBC <> nil then
    ODBC.fQueryList.Add( @ Self );
  fColumns := NewStrListEx;
end;
{$ELSE}
constructor TODBCQuery.Create;
begin
  inherited;
  SQL := TSQLStringList.Create;
  SQL.FOwnerODBCQuery := Self;
  ODBC := DefaultODBCDataSource;
  if ODBC <> nil then
    ODBC.fQueryList.Add( Self );
  fColumns := TStringList.Create;
end;
{$ENDIF}

destructor TODBCQuery.Destroy;
begin
  Close;
  if ODBC <> nil then
    ODBC.fQueryList.Remove( {$IFDEF KOL} @ {$ENDIF} Self );
  SQL.Free;
  ClearColumns;
  fColumns.Free;
  inherited;
end;

function TODBCQuery.ExecSQL: Boolean;
var R: Word;
    Txt: String;
begin
  Result := FALSE;
  Error := 'Not connected';
  if  not ODBC.Connected then Exit;
  ClearColumns;
  if fHandle = 0 then
  begin
    R := SQLAllocHandle( SQL_HANDLE_STMT, ODBC.fConnHandle, @ fHandle );
    if R > 1 then
    begin
      Error := '   :' + Copy( SQL.Text, 1, 100 );
      {$IFDEF KOL_EXCEPTIONS}
      raise EODBC.CreateConnError( ODBC.fConnHandle, R, Error );
      {$ELSE}
      Exit;
      {$ENDIF}
    end;
  end;
  Txt := SQL.Text;

  R := SQLExecDirect( fHandle, PChar( Txt ), Length( Txt ) );
  if (R > 1) and (R <> 100) then
  begin
    Error := '   :' + Copy( Txt, 1, 100 );
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.CreateStmtError( fHandle, R, Error )
    {$ELSE}
    Exit;
    {$ENDIF}
  end
  else Result := TRUE;
end;

function TODBCQuery.FindField(const S: String): Integer;
var i: Integer;
begin
  Result := -1;
  for i := 0 to fColumns.Count-1 do
    if AnsiCompareText( fColumns{$IFDEF KOL}.Items{$ENDIF}[ i ], S ) = 0 then
    begin
      Result := i;
      break;
    end;
end;

procedure TODBCQuery.First;
var R: Word;
begin
  fEof := FALSE;
  fBof := FALSE;
  R := SQLFetchScroll( fHandle, SQL_FETCH_FIRST, 0 );
  if R > 1 then
  begin
    fEof := TRUE;
    fBof := TRUE;
  end;
end;

function TODBCQuery.GetColCount: Integer;
begin
  Result := fColumns.Count;
end;

function TODBCQuery.GetDField(const S: String): TDateTime;
var i: Integer;
begin
  i := ColumnByNameMustBe( S );
  Result := 0;
  if i >= 0 then
    Result := DFieldN[ i ];
end;

function TODBCQuery.GetDFieldN(N: Integer): TDateTime;
var FD: PFieldDescription;
    s: String;
begin
  Result := 0;
  Error := '';
  FD := Pointer( fColumns.Objects[ N ] );
  if FD.StrLenInd^ <> SQL_NULL_DATA then
  CASE FD.DataType OF
  {SQL_NUMERIC:
    Result := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );}
  SQL_DECIMAL:
    begin
      // todo
      asm
        int 3
      end;
    end;
  SQL_CHAR,
  SQL_VARCHAR:
    begin
      TRY
        Result := StrToFloat( Copy( PChar( FD.BindPtr ), 1, FD.StrLenInd^ ) );
      EXCEPT
        raise; //   
      END;
    end;
  SQL_WVARCHAR:
    begin
      TRY
        s := PWideChar( FD.BindPtr )^;
        Result := StrToFloat( s );
      EXCEPT
        raise;
      END;
    end;
  SQL_INTEGER:
    Result := PInteger( FD.BindPtr )^;
  SQL_SMALLINT:
    Result := PSmallInt( FD.BindPtr )^;
  SQL_DOUBLE,
  SQL_FLOAT:
    Result := PDouble( FD.BindPtr )^;
  SQL_REAL:
    Result := PSingle( FD.BindPtr )^;
  {SQL_DATETIME:
    Result := PDateTime( FD.BindPtr )^;}
  SQL_TYPE_TIMESTAMP:
    Result := SqlDateTimeStampToDateTime( FD.BindPtr );
  else
      begin
        Error := '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF}[ N ];
        {$IFDEF KOL_EXCEPTIONS}
        raise EODBC.Create( {$IFDEF KOL}e_custom,{$ENDIF} Error );
        {$ELSE}
        {$ENDIF}
      end;
  END;
end;

function TODBCQuery.GetFField(const S: String): Extended;
var i: Integer;
begin
  i := ColumnByNameMustBe( S );
  Result := 0;
  if i >= 0 then
    Result := FFieldN[ i ];
end;

function TODBCQuery.GetFFieldN(N: Integer): Extended;
var FD: PFieldDescription;
    s: String;
begin
  Result := 0;
  Error := '';
  FD := Pointer( fColumns.Objects[ N ] );
  if FD.StrLenInd^ <> SQL_NULL_DATA then
  CASE FD.DataType OF
  {SQL_NUMERIC:
    Result := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );}
  SQL_DECIMAL:
    begin
      // todo
      asm
        int 3
      end;
    end;
  SQL_CHAR,
  SQL_VARCHAR:
    begin
      TRY
        Result := StrToFloat( Copy( PChar( FD.BindPtr ), 1, FD.StrLenInd^ ) );
      EXCEPT
        raise; //   
      END;
    end;
  SQL_WVARCHAR:
    begin
      TRY
        s := PWideChar( FD.BindPtr )^;
        Result := StrToFloat( s );
      EXCEPT
        raise;
      END;
    end;
  SQL_INTEGER:
    Result := PInteger( FD.BindPtr )^;
  SQL_SMALLINT:
    Result := PSmallInt( FD.BindPtr )^;
  SQL_DOUBLE,
  SQL_FLOAT:
    Result := PDouble( FD.BindPtr )^;
  SQL_REAL:
    Result := PSingle( FD.BindPtr )^;
  SQL_TYPE_TIMESTAMP:
    Result := SqlDateTimeStampToDateTime( FD.BindPtr );
  else
    Error := '     ' + fColumns{$IFDEF KOL}.Items{$ENDIF}[ N ];
    {$IFDEF KOL_EXCEPTIONS}
    raise EODBC.Create( {$IFDEF KOL}e_custom,{$ENDIF} Error );
    {$ELSE}
    {$ENDIF KOL_EXCEPTIONS}
  END;
end;

function TODBCQuery.GetFieldNames(Idx: Integer): String;
begin
  Result := fColumns{$IFDEF KOL}.Items{$ENDIF}[ Idx ];
end;

function TODBCQuery.GetFieldPrecision(Idx: Integer): Integer;
var FD: PFieldDescription;
begin
  FD := Pointer( fColumns.Objects[ Idx ] );
  Result := FD.DecimalDigits;
end;

function TODBCQuery.GetFieldTypes(Idx: Integer): TFieldType;
var FD: PFieldDescription;
begin
  Error := '';
  {ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
    ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
    ftBytes, ftVarBytes, ftAutoInc, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
    ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftFixedChar, ftWideString,
    ftLargeint, ftADT, ftArray, ftReference, ftDataSet, ftOraBlob, ftOraClob,
    ftVariant, ftInterface, ftIDispatch, ftGuid}
  FD := Pointer( fColumns.Objects[ Idx ] );
  CASE FD.DataType OF
  SQL_NUMERIC: Result := ftFloat;
  SQL_DECIMAL: Result := ftBCD;
  SQL_CHAR   ,
  SQL_VARCHAR: Result := ftString;
  SQL_WVARCHAR:Result := ftWideString;
  SQL_INTEGER: Result := ftInteger;
  SQL_SMALLINT:Result := ftSmallint;
  SQL_DOUBLE ,
  SQL_FLOAT  ,
  SQL_REAL   : Result := ftFloat;
  SQL_TYPE_TIMESTAMP:Result := ftDateTime;
  else
    begin
        Error := '     ' +
              fColumns{$IFDEF KOL}.Items{$ENDIF}[ Idx ];
        {$IFDEF KOL_EXCEPTIONS}
        raise EODBC.Create( {$IFDEF KOL}e_custom,{$ENDIF} Error );
        {$ELSE}
        Result := ftUnknown;
        {$ENDIF KOL_EXCEPTIONS}
    end;
  END;
end;

function TODBCQuery.GetIField(const S: String): Integer;
var i: Integer;
begin
  i := ColumnByNameMustBe( S );
  Result := 0;
  if i >= 0 then
    Result := IFieldN[ i ];
end;

function TODBCQuery.GetIFieldN(N: Integer): Integer;
var FD: PFieldDescription;
    s: String;
begin
  Result := 0;
  FD := Pointer( fColumns.Objects[ N ] );
  if FD.StrLenInd^ <> SQL_NULL_DATA then
  CASE FD.DataType OF
  {SQL_NUMERIC:
    Result := Trunc( NumericToExtended( PSqlNumeric( FD.BindPtr )^ ) );}
  SQL_DECIMAL:
    begin
      // todo
      asm
        int 3
      end;
    end;
  SQL_CHAR,
  SQL_VARCHAR:
    begin
      TRY
        Result := StrToInt( Copy( PChar( FD.BindPtr ), 1, FD.StrLenInd^ ) );
      EXCEPT
        raise; //   
      END;
    end;
  SQL_WVARCHAR:
    begin
      TRY
        s := PWideChar( FD.BindPtr )^;
        Result := StrToInt( s );
      EXCEPT
        raise;
      END;
    end;
  SQL_INTEGER:
    Result := PInteger( FD.BindPtr )^;
  SQL_SMALLINT:
    Result := PSmallInt( FD.BindPtr )^;
  SQL_DOUBLE,
  SQL_FLOAT:
    Result := Trunc( PDouble( FD.BindPtr )^ );
  SQL_REAL:
    Result := Trunc( PSingle( FD.BindPtr )^ );
  SQL_TYPE_TIMESTAMP:
    Result := Trunc( SqlDateTimeStampToDateTime( FD.BindPtr ) );
  else
    begin
        Error := '     ' +
              fColumns{$IFDEF KOL}.Items{$ENDIF}[ N ];
        {$IFDEF KOL_EXCEPTIONS}
        raise EODBC.Create( {$IFDEF KOL}e_custom,{$ENDIF} Error );
        {$ELSE}
        {$ENDIF KOL_EXCEPTIONS}
    end;
  END;
end;

function TODBCQuery.GetNullField(const S: String): Boolean;
var i: Integer;
begin
  i := ColumnByNameMustBe( S );
  Result := TRUE;
  if i >= 0 then
    Result := NullFieldN[ i ];
end;

function TODBCQuery.GetNullFieldN(N: Integer): Boolean;
var FD: PFieldDescription;
begin
  FD := Pointer( fColumns.Objects[ N ] );
  Result := FD.StrLenInd^ = SQL_NULL_DATA;
end;

function TODBCQuery.GetSField(const S: String): String;
var i: Integer;
begin
  i := ColumnByNameMustBe( S );
  Result := '';
  if i >= 0 then
    Result := SFieldN[ i ];
end;

function TODBCQuery.GetSFieldN(N: Integer): String;
var FD: PFieldDescription;
begin
  FD := Pointer( fColumns.Objects[ N ] );
  if FD.StrLenInd^ = SQL_NULL_DATA then
    Result := ''
  else
  CASE FD.DataType OF
  (*
  SQL_NUMERIC:
    begin
      E := NumericToExtended( PSqlNumeric( FD.BindPtr )^ );
      Result :=
           FloatToStr( E );
    end;
      Numeric   SQL_DOUBLE !
  *)
  SQL_DECIMAL,
  SQL_CHAR,
  SQL_VARCHAR:
    SetString( Result, PChar( FD.BindPtr ), FD.StrLenInd^ );
  SQL_WVARCHAR:
    begin
      Result := PWideChar( FD.BindPtr )^;
    end;
  SQL_INTEGER:
    Result := IntToStr( PInteger( FD.BindPtr )^ );
  SQL_SMALLINT:
    Result := IntToStr( PSmallInt( FD.BindPtr )^ );
  SQL_DOUBLE,
  SQL_FLOAT:
    Result :=
      {$IFDEF KOL}
      Double2Str( PDouble( FD.BindPtr )^ );
      {$ELSE}
      FloatToStr( PDouble( FD.BindPtr )^ );
      {$ENDIF}
  SQL_REAL:
    Result :=
      {$IFDEF KOL}
      Double2Str( PSingle( FD.BindPtr )^ );
      {$ELSE}
      FloatToStr( PSingle( FD.BindPtr )^ );
      {$ENDIF}
  SQL_TYPE_TIMESTAMP:
    begin
      {$IFDEF KOL}
      Result := DateTime2StrShort( SqlDateTimeStampToDateTime( FD.BindPtr ) );
      {$ELSE}
      Result := FormatDateTime( 'dd.mm.yyyy',
             SqlDateTimeStampToDateTime( FD.BindPtr ) );
      {$ENDIF}
    end;
  END;
end;

procedure TODBCQuery.InternalOpenResultset;
  function FieldSize( FD: PFieldDescription ): Integer;
  begin
    CASE FD.DataType OF
    SQL_NUMERIC: Result := 8; // 19;
    SQL_DECIMAL: Result := 8;
    //SQL_CHAR,
    //SQL_VARCHAR:
    //SQL_WVARCHAR:
    SQL_INTEGER: Result := 4;
    SQL_SMALLINT:Result := 2;
    SQL_DOUBLE:  Result := 8;
    SQL_FLOAT:   Result := 8;
    SQL_REAL:    Result := 10;
    SQL_DATE, SQL_TIME,
    SQL_TYPE_TIMESTAMP:Result := 16;
    else Result := FD.ColumnSize + 1;
    END;
    //Result := 256;
  end;
var i, L, L1: Integer;
    FD: PFieldDescription;
    ColName: array[ 0..4095 ] of Char;
    ColLen: SmallInt;
    s: String;
    R: Word;
    P: PByte;
    t: SmallInt;
begin
  Error := '';
  L := 0;
  i := fColCount;
  ClearColumns;
  fColCount := i;
  for i := 0 to fColCount-1 do
  begin
    FD := AllocMem( Sizeof( FD^ ) );
    SQLDescribeCol( fHandle, i+1, ColName, Sizeof( ColName ), @ ColLen,
      @ FD.DataType, @ FD.ColumnSize, @ FD.DecimalDigits, @ FD.Nullable );
    L1 := FieldSize( FD );
    SetString( s, ColName, ColLen );
    Inc( L, L1 + Sizeof( Integer ) );
    fColumns.AddObject( s, {$IFDEF KOL} DWORD(FD) {$ELSE} Pointer(FD) {$ENDIF} );
  end;
  //SQLSetStmtAttr( fHandle, SQL_ATTR_ROW_BIND_TYPE, Pointer( L ), 0 );
  if  fBindBuffer <> nil then
  asm
    nop
  end;
  fBindBuffer := AllocMem( L );
  P := fBindBuffer;
  for i := 0 to fColCount-1 do
  begin
    FD := Pointer( fColumns.Objects[ i ] );
    FD.BindPtr := P;
    t := FD.DataType;
    L := FieldSize( FD );
    CASE t OF
    SQL_VARCHAR: t := SQL_CHAR;
    SQL_NUMERIC: t := SQL_DOUBLE;
    SQL_FLOAT  : t := SQL_DOUBLE;
    SQL_DECIMAL: t := SQL_DOUBLE;
    SQL_DATE, SQL_TIME,
    SQL_TYPE_TIMESTAMP:t := SQL_TYPE_TIMESTAMP;
    END;
    FD.DataType := t;
    Inc( P, L );
    FD.StrLenInd := Pointer( P );
    Inc( P, Sizeof( Integer ) );
    R := SQLBindCol( fHandle, i+1, t, FD.BindPtr, L, FD.StrLenInd );
    if R > 1 then
    begin
      Error := '    ' +
        fColumns{$IFDEF KOL}.Items{$ENDIF}[ i ];
      {$IFDEF KOL_EXCEPTIONS}
      raise EODBC.CreateStmtError( fHandle, R, Error );
      {$ELSE}
      {$ENDIF KOL_EXCEPTIONS}
    end;
  end;
  Next;
  //First;
end;

procedure TODBCQuery.Last;
var R: Word;
begin
  fEof := FALSE;
  fBof := FALSE;
  R := SQLFetchScroll( fHandle, SQL_FETCH_LAST, 0 );
  if R > 1 then
  begin
    fEof := TRUE;
    fBof := TRUE;
  end;
end;

function TODBCQuery.More: Boolean;
var R: Word;
begin
  Result := FALSE;
  Error := '';
  ClearColumns;
  if fHandle = 0 then Exit;
  REPEAT
    R := SQLMoreResults( fHandle );
    if R > 1 then
    begin
      if R = 100 then Exit;
      {$IFDEF KOL_EXCEPTIONS}
      Error := '  More';
      raise EODBC.CreateStmtError( fHandle, R, Error );
      {$ELSE}
      Exit;
      {$ENDIF KOL_EXCEPTIONS}
    end;
    if R = 0 then
      SQLNumResultCols( fHandle, @ fColCount );
  UNTIL fColCount > 0;
  Result := fColCount > 0;
  if Result then
    InternalOpenResultset;
end;

procedure TODBCQuery.Next;
var R: Word;
begin
  fEof := FALSE;
  R := SQLFetchScroll( fHandle, SQL_FETCH_NEXT, 0 );
  if R > 1 then
    fEof := TRUE
  else
    fBof := FALSE;
end;

function TODBCQuery.Open: Boolean;
begin
  //if Assigned( BeforeOpen ) then
  //  BeforeOpen( nil );
  Result := ExecSQL;
  if  Result then
  begin
      SQLNumResultCols( fHandle, @ fColCount );
      if fColCount = 0 then
        More
      else
        InternalOpenResultset;
  end;
end;

procedure TODBCQuery.Prev;
var R: Word;
begin
  fBof := FALSE;
  R := SQLFetchScroll( fHandle, SQL_FETCH_PRIOR, 0 );
  if R > 1 then
    fBof := TRUE
  else
    fEof := FALSE;
end;

{ TSQLStringList }

{$IFDEF KOL}
{$ELSE}
procedure TSQLStringList.Changed;
begin
  inherited;
  if Assigned( FOwnerODBCQuery ) then
    FOwnerODBCQuery.Close;
end;
{$ENDIF}

end.




